'' =================================================================================================
''
''   File....... jm_xbee_remote_node.spin
''   Purpose.... Smart emote node using XBee coms
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2016 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 
''   Updated.... 22 SEP 2017
''
'' =================================================================================================

{
  Jon's XBee setup using XCTU

  -- start with default configuration
  -- change baud (BD) to 57K
  -- change API enabled (AP) to 1 (no escapes)
  -- change DIO7 (D7) to disabled
  -- change DIO5 (D5) to disabled 
}


con { timing }

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                          ' use 5MHz crystal
                                                                 
  CLK_FREQ = _xinfreq * (_clkmode >> 6)                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us

  TRM_BAUD = 115_200                                            ' baud rate for terminal
  XB_BAUD  = xbee#BR_57K6                                       ' actual xbee baud

                                                                  
con { pab io pins }                                               
                                                                  
  RX1     = 31                                                  ' programming / terminal
  TX1     = 30                                      
                                                    
  SDA1    = 29                                                  ' eeprom / i2c
  SCL1    = 28                                                   
                                                                 
  AUD_RT  = 27                                                  ' audio (d/a) outputs
  AUD_LF  = 26                                                   
                                                                 
  LED_27  = 27                                                  ' leds (on audio outputs)
  LED_26  = 26                                                   
                                                                 
  SD_CS   = 25                                                  ' usd card
  SD_DI   = 24                                                   
  SD_CLK  = 23                                                   
  SD_DO   = 22                                                   
                                                                 
  ADC_CS  = 21                                                  ' adc input
  ADC_CLK = 20                                                   
  ADC_DO  = 19                                                   
  ADC_DI  = 18                                                   
                                                                 
  SVO_6   = 17                                                  ' servo headers
  SVO_5   = 16                                                   
  SVO_4   = 15                                                   
  SVO_3   = 14                                                   
  SVO_2   = 13                                                   
  SVO_1   = 12


con { app io }

  RX2     =  7                                                  ' xbee
  TX2     =  6    


con

  BASE_ID = $BA5E                                               ' "BASE"
                                                   
  REMOTE  = $52 << 8                                            ' $52 = "R" (for remote)
  STATION = $00                                                 ' station #
  MY_ID   = REMOTE | STATION                                    ' full node address

  
  ' commands

  RD_DIGITAL = $D0 { pin as byte }
  WR_DIGITAL = $D1 { pin as byte, value as byte [bit 0] }

  RD_ANALOG  = $A0 { channel as byte [0..3] }
  
  RD_TIME    = $70
  WR_TIME    = $71 { millis as long }


con

  DEBUG = true

  
obj                                                              
                                                                 
' main                                                          ' * master spin cog
  time       : "jm_time_80"                                     '   delays/timing
  io         : "jm_io"                                          '   basic io
  adc        : "jm_adc124s021"                                  '   for onboard adc
  xbee       : "jm_xbee_api1"                                   ' * XBee via API1 interface
' background                                                    ' * background ms timer / processes
  term : "jm_fullduplexserial"                                  ' * serial for terminal (debugging)
                                                                 
' * uses cog when loaded  


var

  byte  packet[16]                                              ' to simplify packet access


dat

  Response              byte    $7E                             ' header
                        byte    $00, $0A                        ' length (full packet - 4)
                        byte    $01                             ' tx reqest (16-bit address)
                        byte    $00                             ' no id required (no auto-response frame)
  Destination           byte    $BA, $5E                        ' destination is BASE (DL, 16-bit)
                        byte    $01                             ' disable ack
  Command               byte    $00                             ' command
  RespData              byte    $00, $00, $00, $00              ' data (Little Endian)
  Checksum              byte    $00                             ' checksum
                                                                 

pub main | id, len                                      
                                                                 
  setup

  ' If debugging, open PST, click Enable, then press a key

  if (DEBUG == true)
    term.rxflush                                                ' clear kbd buffer
    term.rx                                                     ' press a key
    term.tx(term#CLS)                                           ' clear the terminal

  id := xbee.check_radio                                        ' look for radio
  if (id < 0)
    node_error                                                  ' no radio detected
  else
    if (id <> MY_ID)                                            ' ID already set?
      set_my_address                                            '  no, set it

  if (DEBUG == true)
    term.str(string("Ready for packets.", 13, 13))

  repeat
    len := xbee.get_rx_frame(0)                                 ' check for frame waiting
    if (len == 14)                                              ' if present and correct size
      bytemove(@packet, xbee.rx_addr, len)                      ' copy for easy access
      if (packet[3] == $81)                                     ' RX packet
        if (DEBUG == true)
          show_rx_frame(len) 
        process_frame(len)


pub show_rx_frame(len) : idx

'' Display contents of XBee frame
'' -- used when developing/debugging

  if (DEBUG <> true)
    return

  term.str(string("Bytes = "))
  term.dec(len)
  term.tx(13)

  repeat len 
    if (idx > 0) and ((idx // 8) == 0)                          ' 8 bytes per line
      term.tx(13)
    term.tx("$")
    term.hex(packet[idx], 2)                                    ' display as hex
    term.tx(" ")
    ++idx

  term.txn(13, 2)
      

pub process_frame(len) | cmd

  cmd := packet[8]                                              ' extract command   

  if (DEBUG == true)
    term.str(string("Command = "))
    term.hex(cmd, 2)
    term.tx(13)

  case cmd
    $70   : read_millis
    $71   : write_millis
    $A0   : read_analog
    $D0   : read_digital
    $D1   : write_digital
    other : invalid_command


pub invalid_command

  if (DEBUG == true)
    term.str(string("--> Invalid"))
    term.txn(13, 2)


pub bad_global_command 

  if (DEBUG == true)
    term.str(string("--> Global address not supported"))
    term.txn(13, 2)


pub read_millis                                                 ' command $70

'' Read milliseconds timer from this node

  if (DEBUG == true) 
    term.str(string("Read millis timer", 13, "--> "))
    term.dec(millis)
    term.txn(13, 2)

  Command := $70                                                ' echo command
  bytemove(@RespData, @millis, 4)                               ' return millis
  send_response 


pub write_millis : src | ms                                     ' command $71

'' Write long value in packet to millis timer
'' -- will only accept this message from BASE node
'' -- no reponse

  bytemove(@src, @packet[4], 2)                                 ' get message source
  if (src <> $BA5E)                                             ' if not from base node
    return                                                      '  do not allow timer change

  bytemove(@ms, @packet[9], 4)                                  ' extract time from packet
  millis := ms                                                  ' update timer

  if (DEBUG == true) 
    term.str(string("Write millis timer", 13, "--> "))
    term.dec(ms)
    term.txn(13, 2)


pub read_analog | ch                                            ' command $A0  

'' Read analog channel (0..3) from this node
'' -- return 0..4095 for good ch, -1 for bad ch

  ch := packet[9]                                               ' ch is LSB of data

  if (ch =< 3)
    result := adc.read(ch)                                      ' read the ADC channel
  else
    result := -1                                                ' return error code

  if (DEBUG == true)
    term.str(string("Read analog channel "))
    term.dec(ch)
    term.tx(13)
    term.str(string("--> "))
    term.dec(result)
    term.txn(13, 2)  

  Command := $A0                                                ' echo command
  bytemove(@RespData, @result, 4)                               ' move value into packet
  send_response
  

pub read_digital | pin                                          ' command $D0

'' Read digital input from this node
'' -- return 1 (high), 0 (low), -1 (bad pin)

  pin := packet[9]                                              ' pin is LSB of data

  case pin
    00..05 : result := ina[pin]                                 ' good pin
    08..15 : result := ina[pin]                                 ' good pin
    26..27 : result := ina[pin]                                 ' check LED states
    other  : result := -1                                       ' bad pin

  if (DEBUG == true)
    term.str(string("Read digital input "))
    term.dec(pin)
    term.tx(13)
    term.str(string("--> "))
    term.dec(result)
    term.txn(13, 2)

  Command := $D0                                                ' echo command
  bytemove(@RespData, @result, 4)                               ' move value into packet
  send_response


pub write_digital | pin, level                                  ' command $D1

'' Write to available IO pin
'' -- no response

  pin := packet[9]                                              ' pin is LSB of data

 level := packet[10]
  if (level)                                                    ' convert non-zero to 1
    level := 1

  case pin
    00..05 : io.set(pin, level)                                 ' good pin
    08..15 : io.set(pin, level)                                 ' good pin
    26..27 : io.set(pin, level)                                 ' good pin (PAB LEDs)      
    other  : level := -1                                        ' bad pin

  if (DEBUG == true)
    term.str(string("Write digital output "))
    term.dec(pin)
    term.tx(13)
    term.str(string("--> "))
    term.dec(level)
    term.txn(13, 2)


pub send_response

  Checksum := xbee.calc_checksum(@Response[3], 10)              ' add checksum

  xbee.tx_frame(@Response, 14)                                  ' send the packet


pub set_my_address | addr, check

'' Set MY address in XBee

  if (DEBUG == true)
    term.str(string("Configuring radio.", 13))
     
  check := xbee.set_and_verify(string("MY"), MY_ID, 2)          ' set and verify node ID
  time.pause(25)
  xbee.flush

  if (check == true)
    return 

  node_error


pub node_error 

  if (DEBUG == true)
    term.str(string("--> Problem accessing radio.", 13, 13))     

  ' error flash   
    
  repeat                                                      
    io.high(26)
    io.low(27)
    time.pause(250)
    io.low(26)
    io.high(27)
    time.pause(250)


pub setup                                                        
                                                                 
  time.start                                                    ' setup for timing/delays
                                                                 
  io.start(0, 0)                                                ' clear pins (this cog)

  adc.start(ADC_CS, ADC_CLK, ADC_DI, ADC_DO)                    ' start adc driver  

  xbee.startx(RX2, TX2, XB_BAUD)                                ' start serial for xbee *

  bcog := cognew(background, @bstack) + 1                       ' start background timer *

  if (DEBUG == true)                                            
    term.start(RX1, TX1, %0000, TRM_BAUD)                       ' start serial for terminal *


con { background }


var
                                                                     
  long  bcog                                                        
  long  bstack[16]                                                   
                                                                     
  long  millis


pri background : t

  millis := 0

  t := cnt
  repeat
    waitcnt(t += MS_001)
    ++millis
    

dat { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}